/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include <math.h>
#include "Source.h"
#include <string>

#include "SourceLocationWindow.h"

using namespace std;
using namespace FD;

class DisplayLocalization;

DECLARE_NODE(DisplayLocalization);
/*Node
 *
 * @name DisplayLocalization
 * @category HARK:Localization
 * @description Display sound streams in the time-azimuth map.
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description The same as input.
 *
 * @parameter_name WINDOW_NAME
 * @parameter_type string
 * @parameter_value Source Location
 * @parameter_description Window name of the time-azimuth map [default: Window name]
 *
 * @parameter_name WINDOW_LENGTH
 * @parameter_type int
 * @parameter_value 1000
 * @parameter_description Window length to show at the same time [sample]
 * 
 END*/

class DisplayLocalization : public BufferedNode {
  int sourcesID;
  int outputID;
  string window_name;

public:
  DisplayLocalization(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    sourcesID = addInput("SOURCES");
    outputID = addOutput("OUTPUT");

    int windowLength = dereference_cast<int>(parameters.get("WINDOW_LENGTH"));
    init_time_length( windowLength );
    window_name = object_cast<String>(parameters.get("WINDOW_NAME"));

    inputsCache[sourcesID].lookBack = windowLength;
    inOrder = true;
  }

  void initialize(void)
  {
    this->BufferedNode::initialize();

    cout << "initialize\n";
    gdk_threads_enter();
    init_source_location_window(window_name);
    init_start_time();
    draw_plot_area();
    update_window();
    gdk_threads_leave();
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    RCPtr<Vector<ObjectRef> > pSources = getInput(sourcesID, count);
    const Vector<ObjectRef>& sources = *pSources;

    out[count] = pSources;

    if (get_state_alive() && count % 100 == 0) {
      gdk_threads_enter();
      if (source_location_window_is_valid()) {
        update_time(count);
        draw_plot_area();
        plot(count);
        draw_cursor(count);
        update_window();
      }
      gtk_main_iteration_do(FALSE);
      gdk_flush();
      gdk_threads_leave();
    }
  }

  void plot(int count) {
    int start_time = get_start_time();
    int time_length = get_time_length();
    map<int, vector<GdkPoint> > lines;
    int angle;
    GdkPoint pos;

    for (int t = start_time; t <= count; t++) {
      RCPtr<Vector<ObjectRef> > sources = getInput(sourcesID, t);
      for (int i = 0; i < sources->size(); i++) {
        RCPtr<Source> src = (*sources)[i];
        angle = (int)(180 / M_PI * atan2(src->x[1], src->x[0]));
        pos.x = get_x_in_plot_area(t);
        pos.y = get_y_in_plot_area(angle);
        lines[src->id].push_back(pos);
      }
    }

    map<int, vector<GdkPoint> >::iterator it = lines.begin();
    for (; it != lines.end(); ++it) {
      draw_lines(&it->second[0], it->second.size(), it->first);
    }
  }
};
